package cn.com.qmth.examcloud.support.fss.impl;

import cn.com.qmth.examcloud.commons.exception.StatusException;
import cn.com.qmth.examcloud.support.fss.FssHelper;
import cn.com.qmth.examcloud.support.fss.FssProperty;
import cn.com.qmth.examcloud.support.fss.FssService;
import cn.com.qmth.examcloud.support.fss.model.FssFileInfo;
import cn.com.qmth.examcloud.support.fss.model.FssSignInfo;
import cn.com.qmth.examcloud.support.fss.model.FssType;
import com.qcloud.cos.COSClient;
import com.qcloud.cos.ClientConfig;
import com.qcloud.cos.auth.BasicCOSCredentials;
import com.qcloud.cos.auth.COSCredentials;
import com.qcloud.cos.http.HttpProtocol;
import com.qcloud.cos.model.COSObject;
import com.qcloud.cos.model.GetObjectRequest;
import com.qcloud.cos.model.ObjectMetadata;
import com.qcloud.cos.model.PutObjectRequest;
import com.qcloud.cos.region.Region;
import org.apache.commons.lang3.StringUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.InputStream;

public class TencentCosService implements FssService {

    private static final Logger log = LoggerFactory.getLogger(TencentCosService.class);

    @Override
    public FssFileInfo writeFile(String filePath, File file, String md5) {
        if (file == null) {
            throw new StatusException("文件不能为空！");
        }
        if (!file.isFile()) {
            throw new StatusException("文件类型错误！");
        }
        if (file.length() == 0) {
            throw new StatusException("文件大小不能为0！");
        }

        if (StringUtils.isNotEmpty(md5)) {
            String realMd5 = FssHelper.getFileMD5(file);
            if (!md5.equals(realMd5)) {
                log.warn("filePath:{} realMD5:{} MD5:{}", filePath, realMd5, md5);
                throw new StatusException("文件MD5值校验不通过！");
            }
        }

        filePath = this.fixCosFilePath(filePath);
        COSClient client = this.getClient();

        try {
            PutObjectRequest objectRequest = new PutObjectRequest(FssProperty.FSS_BUCKET, filePath, file);
            client.putObject(objectRequest);
        } catch (Exception e) {
            log.error("文件上传失败！filePath:{} err:{}", filePath, e.getMessage());
            throw new StatusException("文件上传失败！");
        } finally {
            try {
                client.shutdown();
            } catch (Exception e) {
                // ignore
            }
        }

        filePath = FssProperty.FSS_SEPARATOR + filePath;
        FssFileInfo result = new FssFileInfo();
        result.setFileName(FssHelper.getFileName(filePath));
        result.setFilePath(filePath);
        result.setFileUrl(FssProperty.FSS_URL_PREFIX + filePath);
        return result;
    }

    @Override
    public FssFileInfo writeFile(String filePath, byte[] bytes, String md5) {
        if (StringUtils.isNotEmpty(md5)) {
            String realMd5 = FssHelper.getFileMD5(bytes);
            if (!md5.equals(realMd5)) {
                log.warn("filePath:{} realMD5:{} MD5:{}", filePath, realMd5, md5);
                throw new StatusException("文件MD5值校验不通过！");
            }
        }

        filePath = this.fixCosFilePath(filePath);
        COSClient client = this.getClient();

        try (ByteArrayInputStream is = new ByteArrayInputStream(bytes);) {
            ObjectMetadata objectMetadata = new ObjectMetadata();
            objectMetadata.setContentLength(bytes.length);
            PutObjectRequest objectRequest = new PutObjectRequest(FssProperty.FSS_BUCKET, filePath, is, objectMetadata);
            client.putObject(objectRequest);
        } catch (Exception e) {
            log.error("文件上传失败！filePath:{} err:{}", filePath, e.getMessage());
            throw new StatusException("文件上传失败！");
        } finally {
            try {
                client.shutdown();
            } catch (Exception e) {
                // ignore
            }
        }

        filePath = FssProperty.FSS_SEPARATOR + filePath;
        FssFileInfo result = new FssFileInfo();
        result.setFileName(FssHelper.getFileName(filePath));
        result.setFilePath(filePath);
        result.setFileUrl(FssProperty.FSS_URL_PREFIX + filePath);
        return result;
    }

    @Override
    public void readFile(String filePath, File toFile) {
        filePath = this.fixCosFilePath(filePath);
        COSClient client = this.getClient();

        try {
            GetObjectRequest request = new GetObjectRequest(FssProperty.FSS_BUCKET, filePath);

            FssHelper.makeDirs(toFile.getParent());
            client.getObject(request, toFile);
        } catch (Exception e) {
            log.error("文件下载失败！filePath:{} err:{}", filePath, e.getMessage());
            throw new StatusException("文件下载失败！");
        } finally {
            try {
                client.shutdown();
            } catch (Exception e) {
                // ignore
            }
        }
    }

    @Override
    public byte[] readFile(String filePath) {
        filePath = this.fixCosFilePath(filePath);
        COSClient client = this.getClient();

        InputStream is = null;
        ByteArrayOutputStream bos = null;
        try {
            COSObject cosObject = client.getObject(FssProperty.FSS_BUCKET, filePath);
            is = cosObject.getObjectContent();
            bos = new ByteArrayOutputStream();

            int len;
            byte[] bytes = new byte[1024];
            while ((len = is.read(bytes)) != -1) {
                bos.write(bytes, 0, len);
            }
            bos.flush();
            cosObject.close();

            return bos.toByteArray();
        } catch (Exception e) {
            log.error("文件下载失败！filePath:{} err:{}", filePath, e.getMessage());
            throw new StatusException("文件下载失败！");
        } finally {
            try {
                if (is != null) {
                    is.close();
                }
                if (bos != null) {
                    bos.close();
                }

                client.shutdown();
            } catch (Exception e) {
                // ignore
            }
        }
    }

    @Override
    public boolean deleteFile(String filePath) {
        try {
            filePath = this.fixCosFilePath(filePath);
            this.getClient().deleteObject(FssProperty.FSS_BUCKET, filePath);
            return true;
        } catch (Exception e) {
            log.error(e.getMessage(), e);
            return false;
        }
    }

    @Override
    public boolean existFile(String filePath) {
        filePath = this.fixCosFilePath(filePath);
        return this.getClient().doesObjectExist(FssProperty.FSS_BUCKET, filePath);
    }

    @Override
    public void refreshFile(String fileUrls) {
        //  控制台 --> 开通云函数服务 --> 设置CDN缓存刷新函数
        // @see https://cloud.tencent.com/document/product/436/45597
    }

    @Override
    public FssSignInfo buildSign(String filePath) {
        FssSignInfo info = new FssSignInfo();
        info.setFsType(FssType.TENCENT_COS.name());
        info.setSignIdentifier(String.valueOf(System.currentTimeMillis()));
        //todo

        return info;
    }

    private COSClient getClient() {
        try {
            Region region = new Region(FssProperty.FSS_REGION_ID);
            ClientConfig clientConfig = new ClientConfig(region);
            clientConfig.setHttpProtocol(HttpProtocol.https);
            clientConfig.setMaxErrorRetry(1);

            // 同地域的内网访问? COS会智能解析到内网IP
            // @see https://cloud.tencent.com/document/product/436/56556

            COSCredentials cred = new BasicCOSCredentials(FssProperty.FSS_ACCESS_KEY_ID,
                    FssProperty.FSS_ACCESS_KEY_SECRET);
            return new COSClient(cred, clientConfig);
        } catch (Exception e) {
            log.error(e.getMessage(), e);
            throw new StatusException("COS客户端初始化失败！");
        }
    }

    private String fixCosFilePath(String filePath) {
        if (StringUtils.isEmpty(filePath)) {
            throw new StatusException("文件存储路径不能为空！");
        }

        // 存储路径统一去掉 / 开头
        if (filePath.startsWith("/")) {
            return filePath.substring(1);
        }

        return filePath;
    }

}
